Attribute VB_Name = "mInterp"
Option Explicit
'Accepts a data_range cell reference which is a table of numbers, where the first row and first column are the data labels
Public Function UDF_BiLinearInterp(ByRef data_range As Excel.Range, ByVal x_interp As Double, ByVal y_interp As Double) As Variant
    Dim data_values() As Variant, row_index As Long, col_index As Long
    Dim x_values() As Double, y_values() As Double, z_values() As Double
    
    If data_range.Columns.Count < 2 Then
        'Need at least two columns (y labels and 1 y value)
        UDF_BiLinearInterp = CVErr(xlErrNull)
        Exit Function
    ElseIf data_range.Rows.Count < 2 Then
        'Need at least two rows (x labels and 1 x value)
        UDF_BiLinearInterp = CVErr(xlErrNull)
        Exit Function
    End If

    On Error GoTo CatchCleanup
    
    'Allocate memory for sub function calls
    ReDim x_values(0 To data_range.Columns.Count - 2)
    ReDim y_values(0 To data_range.Rows.Count - 2)
    ReDim z_values(0 To UBound(y_values), 0 To UBound(x_values))
    
    'Load cell values into memory (fast)
    data_values = data_range.Value
    
    'Split into separate double arrays for sub function
    For row_index = 1 To data_range.Rows.Count
        For col_index = 1 To data_range.Columns.Count
            If row_index = 1 Then
                If col_index = 1 Then
                    'Ignore
                Else
                    'First row = X header labels
                    x_values(col_index - 2) = data_values(row_index, col_index)
                    If col_index >= 3 Then
                        'Check that X values are in ascending order, otherwise return #NUM! error
                        If x_values(col_index - 3) >= x_values(col_index - 2) Then
                            UDF_BiLinearInterp = CVErr(xlErrNum)
                            GoTo CatchCleanup
                        End If
                    End If
                End If
            ElseIf col_index = 1 Then
                'First column = Y header labels
                y_values(row_index - 2) = data_values(row_index, col_index)
                If row_index >= 3 Then
                    'Check that Y values are in ascending order, otherwise return #NUM! error
                    If y_values(row_index - 3) >= y_values(row_index - 2) Then
                        UDF_BiLinearInterp = CVErr(xlErrNum)
                        GoTo CatchCleanup
                    End If
                End If
            Else
                'table value
                z_values(row_index - 2, col_index - 2) = data_values(row_index, col_index)
            End If
        Next
    Next
    
    'Calculate interp
    UDF_BiLinearInterp = BiLinearInterp(x_values, y_values, z_values, x_interp, y_interp)
    
    'Cleanup
CatchCleanup:
    Erase x_values
    Erase y_values
    Erase z_values
    Erase data_values
    On Error GoTo 0
End Function

Public Function UDF_LinearInterp(ByRef x_range As Excel.Range, ByRef y_range As Excel.Range, ByVal x_interp As Double) As Variant
    Dim x_values() As Double, y_values() As Double, range_cell As Range, i As Long
    If x_range.Count <> y_range.Count Then
        UDF_LinearInterp = CVErr(xlErrNull)
        Exit Function
    End If
    ReDim x_values(0 To x_range.Count - 1)
    ReDim y_values(0 To y_range.Count - 1)
    i = 0
    For Each range_cell In x_range.Cells
        x_values(i) = range_cell.Value
        'Ensure X values are ascending order
        If i >= 1 Then
            If x_values(i - 1) >= x_values(i) Then
                UDF_LinearInterp = CVErr(xlErrNum)
                GoTo ExitCleanup
            End If
        End If
        '
        i = i + 1
    Next
    i = 0
    For Each range_cell In y_range.Cells
        y_values(i) = range_cell.Value
        i = i + 1
    Next
    UDF_LinearInterp = LinearInterp(x_values, y_values, x_interp)
    
ExitCleanup:
    Erase x_values
    Erase y_values
    Set range_cell = Nothing
End Function

Public Function BiLinearInterp(ByRef x_values() As Double, ByRef y_values() As Double, ByRef z_values() As Double, ByVal x_interp As Double, ByVal y_interp As Double) As Double
    Dim x_index As Long, y_index As Long
    Dim x_ubound As Long, y_ubound As Long
    Dim x_interp_fixed As Double, y_interp_fixed As Double
    Dim lo_index As Long, hi_index As Long, mid_index As Long
    
    Debug.Assert LBound(x_values) = 0                                   'Ensure X array starts at index 0
    Debug.Assert UBound(x_values) > 0                                   'Ensure X array has more than 1 value
    Debug.Assert LBound(y_values) = 0                                   'Ensure Y array starts at index 0
    Debug.Assert UBound(y_values) > 0                                   'Ensure Y array has more than 1 value
    Debug.Assert LBound(z_values, 1) = 0 And LBound(z_values, 2) = 0    'Ensure Z array starts at index 0
    Debug.Assert UBound(z_values, 1) = UBound(y_values)                 'Ensure Z rows = Y length
    Debug.Assert UBound(z_values, 2) = UBound(x_values)                 'Ensure Z columns = X length
    
    'Limit the interp bounds to max-min of X/Y values
    x_ubound = UBound(x_values)
    x_interp_fixed = IIf(x_interp > x_values(0), x_interp, x_values(0))
    x_interp_fixed = IIf(x_interp_fixed < x_values(x_ubound), x_interp_fixed, x_values(x_ubound))
    y_ubound = UBound(y_values)
    y_interp_fixed = IIf(y_interp > y_values(0), y_interp, y_values(0))
    y_interp_fixed = IIf(y_interp_fixed < y_values(y_ubound), y_interp_fixed, y_values(y_ubound))
    
    'Binary search the X indices
    lo_index = 0
    hi_index = x_ubound
    mid_index = CLng((lo_index + hi_index) / 2)
    
    Do While (hi_index - lo_index) >= 3
        If x_interp_fixed < x_values(mid_index) Then
            hi_index = mid_index
        Else
            lo_index = mid_index
        End If
        mid_index = CLng((lo_index + hi_index) / 2)
    Loop
    If x_interp_fixed < x_values(mid_index) Then
        x_index = lo_index
    Else
        x_index = mid_index
    End If
    
    'Binary search the Y index
    lo_index = 0
    hi_index = y_ubound
    mid_index = CLng((lo_index + hi_index) / 2)
    
    Do While (hi_index - lo_index) >= 3
        If y_interp_fixed < y_values(mid_index) Then
            hi_index = mid_index
        Else
            lo_index = mid_index
        End If
        mid_index = CLng((lo_index + hi_index) / 2)
    Loop
    If y_interp_fixed < y_values(mid_index) Then
        y_index = lo_index
    Else
        y_index = mid_index
    End If
    
    'Compute the interpolated value and return
    BiLinearInterp = _
        ( _
            (z_values(y_index + 1, x_index) - z_values(y_index + 1, x_index + 1)) / _
            (x_values(x_index) - x_values(x_index + 1)) * _
            (x_interp - x_values(x_index)) + _
            z_values(y_index + 1, x_index) - _
            (z_values(y_index, x_index) - z_values(y_index, x_index + 1)) / _
            (x_values(x_index) - x_values(x_index + 1)) * _
            (x_interp - x_values(x_index)) - _
            z_values(y_index, x_index) _
        ) _
        / _
        (y_values(y_index + 1) - y_values(y_index)) * _
        (y_interp - y_values(y_index + 1)) + _
        (z_values(y_index + 1, x_index) - z_values(y_index + 1, x_index + 1)) / _
        (x_values(x_index) - x_values(x_index + 1)) * _
        (x_interp - x_values(x_index)) + _
        z_values(y_index + 1, x_index)
End Function

'Linear Interpolator
Public Function LinearInterp(ByRef x_values() As Double, ByRef y_values() As Double, ByVal x_interp As Double) As Double
    Dim x_index As Long, x_ubound As Long
    Dim x_interp_fixed As Double
    Dim lo_index As Long, hi_index As Long, mid_index As Long
    
    Debug.Assert LBound(x_values) = 0                   'Ensure X array starts at index 0
    Debug.Assert UBound(x_values) > 0                   'Ensure X array has more than 1 value
    Debug.Assert LBound(y_values) = 0                   'Ensure Y array starts at index 0
    Debug.Assert UBound(y_values) > 0                   'Ensure Y array has more than 1 value
    Debug.Assert UBound(x_values) = UBound(y_values)    'Ensure X array length matches Y array length
    
    x_ubound = UBound(x_values)
    x_interp_fixed = IIf(x_interp > x_values(0), x_interp, x_values(0))
    x_interp_fixed = IIf(x_interp_fixed < x_values(x_ubound), x_interp_fixed, x_values(x_ubound))
    
    'Binary search the index
    lo_index = 0
    hi_index = x_ubound
    mid_index = CLng((lo_index + hi_index) / 2)
    
    Do While (hi_index - lo_index) >= 3
        If x_interp_fixed < x_values(mid_index) Then
            hi_index = mid_index
        Else
            lo_index = mid_index
        End If
        mid_index = CLng((lo_index + hi_index) / 2)
    Loop
    If x_interp_fixed < x_values(mid_index) Then
        x_index = lo_index
    Else
        x_index = mid_index
    End If
    
    'Compute the interpolated value and return
    LinearInterp = (x_interp - x_values(x_index)) / (x_values(x_index + 1) - x_values(x_index)) * (y_values(x_index + 1) - y_values(x_index)) + y_values(x_index)
End Function

Private Sub Test_LinearInterp()
    Dim x_values() As Variant, y_values() As Variant
    x_values = Array(0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
    y_values = Array(0, 10, 20, 30, 40, 50, 60, 70, 80, 90)
    
    Debug.Print LinearInterp(CDbl_Array(x_values), CDbl_Array(y_values), -1)
End Sub

Private Sub Test_BiLinearInterp()
    Dim x_values() As Variant, y_values() As Variant, z_values() As Double
    x_values = Array(0, 1)
    y_values = Array(0, 10)
    ReDim z_values(0 To 1, 0 To 1)
    z_values(0, 0) = 0
    z_values(0, 1) = 1
    z_values(1, 0) = 1
    z_values(1, 1) = 0
    Debug.Print BiLinearInterp(CDbl_Array(x_values), CDbl_Array(y_values), z_values, 0.5, 5#)
End Sub

Private Function CDbl_Array(ByRef var_array() As Variant) As Double()
    Dim result() As Double, i As Long
    ReDim result(LBound(var_array) To UBound(var_array))
    For i = LBound(var_array) To UBound(var_array)
        result(i) = var_array(i)
    Next
    CDbl_Array = result
End Function

